/**
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 2009-2012 Bull S.A.S.
 * Contact: jonas-team@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */

package org.ow2.jonas.admin.auditconsole.service;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.Map;
import java.util.regex.PatternSyntaxException;

import javax.management.InstanceNotFoundException;
import javax.management.MBeanException;
import javax.management.MBeanServerConnection;
import javax.management.ObjectName;
import javax.management.ReflectionException;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.ow2.util.auditreport.impl.event.Event;
import org.ow2.util.auditreport.impl.GenericAuditReport;
import org.ow2.util.event.api.EventPriority;
import org.ow2.util.event.api.IEvent;
import org.ow2.util.event.api.IEventListener;
import org.ow2.util.event.api.IEventService;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

import org.apache.felix.ipojo.annotations.Component;
import org.apache.felix.ipojo.annotations.Instantiate;
import org.apache.felix.ipojo.annotations.Invalidate;
import org.apache.felix.ipojo.annotations.Provides;
import org.apache.felix.ipojo.annotations.Requires;
import org.apache.felix.ipojo.annotations.ServiceProperty;
import org.apache.felix.ipojo.annotations.Validate;

import org.ow2.kerneos.core.service.KerneosSimpleService;

import org.ow2.util.log.Log;
import org.ow2.util.log.LogFactory;

/**
 * Listener for audit notifications.
 */
@Component
@Instantiate
@Provides
public class AuditService implements KerneosSimpleService, IEventListener {

    @ServiceProperty(name = KerneosSimpleService.ID, value = "auditService")
    private String id;

    /**
     * Logger of the class
     */
    public static Log logger = LogFactory.getLog(AuditService.class);

    /**
     * Reports waiting to be displayed.
     */
    private List<GenericAuditReport> reportsObjects;

    /**
     * The EZB JMX server.
     */
    private transient MBeanServerConnection mbeanServer = null;

    /**
     * MBean name of the logger.
     */
    private ObjectName loggerMBean = null;

    /**
     * Number of invocations detected.
     */
    private int invocations = 0;

    /**
     * Domain name.
     */
    private String domainName = null;

    private CallTops tops;

    private HealthService health;

    /**
     * eventService to get events
     */
    @Requires
    private IEventService eventService;

    /**
     * Constructor.
     */
    public AuditService() {
    }

    @Validate
    private void start() {
        logger.debug("");
        tops = new CallTops();
        health = new HealthService();
        init();
    }

    @Invalidate
    private void stop() {
        logger.debug("");
    }

    /**
     * Init the notifications listening.
     */
    private void init() {
        reportsObjects = new ArrayList<GenericAuditReport>();
    }

    /**
     * Start the listening.
     */
    public void startListening() {
        getNewRegistrations();
        this.reportsObjects.clear();
    }

    /**
     * @return waiting reports.
     */
    public ArrayList<GenericAuditReport> getReportObjects() {
        ArrayList<GenericAuditReport> tmp = new ArrayList<GenericAuditReport>(reportsObjects);
        this.reportsObjects.clear();
        return tmp;
    }

    /**
     * Start the audit on every auditor component.
     */
    public void startAuditing() {
        // TODO : start jonas auditing services .
    }

    /**
     * Stop the audit on every auditor component.
     */
    public void stopAuditing() {
        // TODO : stop jonas auditing services.
    }

    /**
     * this methode handle a eventService event
     * @param event the event
     */
    public void handle(IEvent event) {
        GenericAuditReport report = ((Event) event).getReport();
        this.reportsObjects.add(report);
        tops.addReport(report);
        health.addReport(report);
        invocations++;
    }

    /**
     * This method registers listener (this) on eventService server.
     */
    public void getNewRegistrations() {
        try {
            if (eventService != null) {
                logger.info("Start listening audit notifications");
                eventService.registerListener(this, "EJB|JAXWS|JNDI|WEB");
            } else {
                logger.info("No eventService: Cannot start listening audit notifications");
            }
        } catch (Exception ex) {
            logger.error("Error at listeners registration : ", ex);
        }
    }

    /**
     * Unregister the event Listener (this).
     */
    public void unregister() {
        logger.info("Stop listening audit notifications");
        try {
            if (eventService != null) {
                eventService.unregisterListener(this);
            }
        } catch (Exception ex) {
            logger.error("Error at listeners registration : ", ex);
        }
    }

    /**
     * Get a Map with data name and data value.
     * @param id the NODEID on which we want to retrieve datas
     * @return the expected map
     * @throws IOException if there is a failure
     * @throws ReflectionException if there is a failure
     * @throws InstanceNotFoundException if there is a failure
     * @throws MBeanException if there is a failure
     */
    @SuppressWarnings("unchecked")
    protected Map<String, Object> getDescriptionNodeId(final String id) throws IOException, ReflectionException,
        InstanceNotFoundException, MBeanException {
        Map<String, Object> map = null;
        if (mbeanServer != null) {
            if (mbeanServer.isRegistered(loggerMBean)) {
                map = (Map<String, Object>) mbeanServer.invoke(loggerMBean, "getDescriptionNodeId", new Object[] {id},
                    new String[] {String.class.getName()});
            }
        }
        // if N/A
        return map;
    }

    /**
     * Gets Node Info for the given node ID.
     * @param nodeID the expected Node ID
     * @return data for the given node
     * @throws IOException if there is a failure
     * @throws ReflectionException if there is a failure
     * @throws InstanceNotFoundException if there is a failure
     * @throws MBeanException if there is a failure
     */
    protected NodeInfo getNodeInfo(final String nodeID) throws IOException, ReflectionException, InstanceNotFoundException,
        MBeanException {
        Map<String, Object> map = getDescriptionNodeId(nodeID);
        NodeInfo nodeInfo = null;
        if (map != null) {
            nodeInfo = new NodeInfo();
            nodeInfo.id = nodeID;
            nodeInfo.date = (Date) map.get("date");
            nodeInfo.name = (String) map.get("name");
            nodeInfo.application = (String) map.get("application");
            nodeInfo.desc = (String) map.get("desc");
            nodeInfo.type = (String) map.get("type");
            nodeInfo.depth = (Integer) map.get("depth");
            nodeInfo.localChildren = (String[]) map.get("children");
            nodeInfo.execTime = (Double) map.get("execTime");
        }
        return nodeInfo;

    }

    /**
     * Gets all available nodes at the root level (entry points)
     * @param keyword a keyword used for the search
     * @param maxResults the maximum results to send
     * @param depth the wanted depth
     * @return the node Info list if there is a failure
     * @throws InstanceNotFoundException if there is a failure
     * @throws MBeanException if there is a failure
     * @throws ReflectionException if there is a failure
     * @throws IOException if there is a failure
     */
    public List<NodeInfo> getTreeAvailableNodes(final String keyword, final int maxResults, final int depth)
        throws InstanceNotFoundException, MBeanException, ReflectionException, IOException {
        List<NodeInfo> nodesInfo = new ArrayList<NodeInfo>();
        String[] nodesID = null;
        // Get list of nodes ID
        if (mbeanServer != null) {
            if (mbeanServer.isRegistered(loggerMBean)) {
                nodesID = (String[]) mbeanServer.invoke(loggerMBean, "getRootNodesID", null, null);
            }
        }
        // Nodes available
        if (nodesID != null) {
            // For each node, get a description
            for (String nodeID : nodesID) {
                // Break when we have the results
                if (nodesInfo.size() >= maxResults) {
                    break;
                }
                NodeInfo nodeInfo = getNodeInfo(nodeID);

                if (nodeInfo != null && nodeInfo.name != null) {
                    // Matching name ?
                    boolean matching = false;
                    try {
                        matching = nodeInfo.name.matches(keyword);
                    } catch (PatternSyntaxException e) {
                        matching = false;
                    }
                    if (!matching) {
                        continue;
                    }

                    if (nodeInfo.depth >= depth) {
                        nodesInfo.add(nodeInfo);
                    }

                }
            }
        }

        return nodesInfo;

    }

    /**
     * Build a new Element for the NODE Graph object.
     * @param document the document used to build the element
     * @param id the id of the node
     * @param nodeInfo the data for this node
     * @return a built node element
     */
    protected Element nodeElement(final Document document, final String id, final NodeInfo nodeInfo) {
        Element nodeElement = document.createElement("Node");
        nodeElement.setAttribute("id", id);
        nodeElement.setAttribute("name", nodeInfo.name);
        nodeElement.setAttribute("desc", nodeInfo.desc);
        nodeElement.setAttribute("nodeColor", "0xFF0000");

        String icon = "leaf";
        if ("EJB".equals(nodeInfo.type)) {
            icon = "10";
        } else if ("EJB_LOCAL".equals(nodeInfo.type)) {
            icon = "11";
        } else if ("EJB_REMOTE".equals(nodeInfo.type)) {
            icon = "12";
        } else if ("HTTP_GET".equals(nodeInfo.type)) {
            icon = "2";
        } else if ("HTTP_POST".equals(nodeInfo.type)) {
            icon = "2";
        } else if ("JAXWS_ENDPOINT".equals(nodeInfo.type)) {
            icon = "30";
        }
        nodeElement.setAttribute("nodeIcon", icon);
        return nodeElement;
    }

    /**
     * Build a new Element for the EDGE Graph object.
     * @param document the document used to build the element
     * @param parentId the Parent id of the node (used to make a link between parent and current ID)
     * @param id the id of the node
     * @param label the label of this edge
     * @param flow the flow of this edge
     * @return a built edge element
     */
    protected Element edgeElement(final Document document, final String parentId, final String id, final String label, final String flow) {
        Element nodeElement = document.createElement("Edge");
        nodeElement.setAttribute("fromID", parentId);
        nodeElement.setAttribute("toID", id);
        nodeElement.setAttribute("edgeLabel", label);
        nodeElement.setAttribute("flow",flow);
        nodeElement.setAttribute("color","0x000000");

        return nodeElement;
    }

    /**
     * Complete the given document by adding XML nodes and edges.
     * @param document the XML document
     * @param root the entry node
     * @param first in order to know if we need to link with a parent or not
     * @param parentID the parent ID
     * @param id the current ID
     * @param nodeInfo the info about the node
     * @throws InstanceNotFoundException if there is a failure
     * @throws ReflectionException if there is a failure
     * @throws MBeanException if there is a failure
     * @throws IOException if there is a failure
     */
    private void buildXML(final Document document, final Element root, final boolean first, final String parentID, final String id, final NodeInfo nodeInfo) throws InstanceNotFoundException, ReflectionException, MBeanException, IOException {

        // Link between parent and ourself
        if (!first) {
            root.appendChild(edgeElement(document, parentID, id, String.valueOf(nodeInfo.execTime) + " ms", "0"));
        }

        // Add the node
        root.appendChild(nodeElement(document, id, nodeInfo));

        String[] children = nodeInfo.localChildren;

        // For each children, add the XML data
        if (children != null) {
            int length = children.length;
            int i = 0;
            for (int c = length -1; c >= 0; c--) {
                String idCurrent = id + "_" + i;
                NodeInfo localNodeInfo = getNodeInfo(children[c]);
                buildXML(document, root, false, id, idCurrent, localNodeInfo);
                i++;
            }
        }

    }


    /**
     * Build an XML document and send it to the client
     * @param nodeID the root node ID
     * @return a document with Nodes and Edges
     * @throws InstanceNotFoundException if there is a failure
     * @throws ReflectionException if there is a failure
     * @throws MBeanException if there is a failure
     * @throws IOException if there is a failure
     */
    public Document getGraphNodeID(final String nodeID) throws InstanceNotFoundException, ReflectionException, MBeanException, IOException {
        // Create builder with factory
        DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
        DocumentBuilder builder = null;
        try {
            builder = factory.newDocumentBuilder();
        } catch (ParserConfigurationException e) {
            throw new IllegalStateException("Cannot build document builder", e);
        }

        // Create Document
        Document document = builder.newDocument();

        // Append root element
        Element root = document.createElement("Graph");
        document.appendChild(root);

        // Add root node


        buildXML(document, root, true, "", nodeID, getNodeInfo(nodeID));

        return document;

    }

    public List<TopMethod> getGlobalTop() {
        return tops.getGlobalTop();
    }

    public List<TopMethod> getEjbTop() {
        return tops.getEjbTop();
    }

    public List<TopMethod> getWebTop() {
        return tops.getWebTop();
    }

    public List<TopMethod> getJndiTop() {
        return tops.getJndiTop();
    }

    public void setLimit(int limit) {
        tops.setTopsNumber(limit);  
    }

    public ServerHealth getServerHealth() {
        return health.getServerHealth();
    }

    /**
     * @return number of invocations.
     */
    public int getInvocations() {
        return invocations;
    }

    /**
     * this method accept events reception on JOnAs admin
     */
    public boolean accept(IEvent event) {
        return true;
    }

    /**
     * to get EventPriority
     */
    public EventPriority getPriority() {

        return EventPriority.ASYNC_NORM;
    }
}